import logging
import os.path
from datetime import datetime
from taichu_storage import StorageInterface
from obs import ObsClient, PutObjectHeader
import mimetypes
from io import BytesIO


class StorageObs(StorageInterface):

    def __init__(self, cfgs=None):
        if cfgs is None:
            cfgs = {}

        self._bucket = cfgs.get('bucket')
        obs_ak = cfgs.get('ak')
        obs_sk = cfgs.get('sk')
        obs_server = cfgs.get('endpoint_url')

        self._client = ObsClient(
            access_key_id=obs_ak,
            secret_access_key=obs_sk,
            server=obs_server
        )
        resp = self._client.listBuckets()
        if resp.status >= 300:
            raise Exception(f'Failed to connect to OBS,errorCode:{resp.errorCode}, errorMessage: {resp.errorMessage}')
        resp = self._client.headBucket(bucketName=self._bucket)
        if resp.status == 404:
            raise Exception(f'Bucket {self._bucket} does not exist')
        print("Successfully connected to OBS!")

    def list_objects(self, key, delimiter='', skip_empty_file=True):
        max_num = 1000
        mark = None
        result = []
        while True:
            resp = self._client.listObjects(bucketName=self._bucket, prefix=key, marker=mark,
                                            max_keys=max_num, delimiter=delimiter)
            if resp.status < 300:
                folders = []
                for folder in resp.body.commonPrefixs:
                    folders.append({
                        'name': folder.Prefix,
                        'is_dir': True,
                        'size': 0,
                        'last_modified': None,
                    })
                result.extend(folders)
                objects = []
                for content in resp.body.contents:
                    if skip_empty_file and content.key.endswith('/'):
                        continue
                    objects.append({
                        'name': content.key,
                        'is_dir': False,
                        'size': content.size,
                        'last_modified': datetime.strptime(content.lastModified, "%Y/%m/%d %H:%M:%S").strftime(
                            "%Y-%m-%d %H:%M:%S"),
                    })
                result.extend(objects)
                if resp.body.is_truncated is True:
                    mark = resp.body.next_marker
                else:
                    break
            else:
                logging.error(f'list_objects error,error_code:{resp.status}, message:{resp.errorMessage}')
                break

        return result

    def get_object(self, key):
        response = self._client.getObject(self._bucket, key, loadStreamInMemory=True)
        if response.status < 300:
            return BytesIO(response.body.buffer)
        return BytesIO()

    def put_object(self, key, content):
        self._client.putContent(self._bucket, key, content=content)

    def upload_file(self, local_filename, key):
        headers = PutObjectHeader()
        content_type = 'text/plain'
        mime_type = mimetypes.guess_type(key)[0]
        if mime_type == 'text/plain':
            mime_type = 'text/plain;charset=utf-8'
        if mime_type:
            content_type = mime_type
        headers.contentType = content_type

        self._client.putFile(self._bucket, key, local_filename, metadata={}, headers=headers)

    def upload_dir(self, local_dir, key):
        local_dir = local_dir.rstrip('/')
        key = key.rstrip('/')
        self.upload_file(local_dir, key)

    def download_file(self, key, local_filename):
        if local_filename.endswith('/'):
            os.makedirs(local_filename, exist_ok=True)
            local_filename = os.path.join(local_filename, key.split('/')[-1])

        resp = self._client.getObject(self._bucket, key, downloadPath=local_filename)
        if not resp.status < 300:
            logging.error(f"key:{key},local_filename:{local_filename}")
            logging.error(f'errorCode:{resp.errorCode}, errorMessage:{resp.errorMessage}')

    def download_dir(self, key, local_dir):
        if local_dir.endswith('/'):
            local_dir = local_dir[:-1]
        objects = self.list_objects(key, skip_empty_file=False)
        for o in objects:
            local_filename = os.path.join(local_dir, o['name'].replace(key, ''))
            if o['is_dir']:
                continue
            if o['name'].endswith('/'):
                os.makedirs(local_filename, exist_ok=True)
                continue

            self.download_file(o['name'], local_filename)

    def copy_object(self, source_key, dest_key):
        resp = self._client.copyObject(self._bucket, source_key, self._bucket, dest_key)
        if not resp.status < 300:
            logging.error(f'copy_object error,error_code:{resp.errorCode}, message:{resp.errorMessage}')

    def copy_dir(self, source_path, dest_path):
        if not dest_path.endswith('/'):
            dest_path = dest_path + '/'
        objects = self.list_objects(source_path)
        for o in objects:
            source_key = o['name']
            dest_key = source_key.replace(source_path, dest_path)
            self.copy_object(source_key, dest_key)

    def create_dir(self, key):
        if not key.endswith('/'):
            key = key + '/'
        self.put_object(None, key)

    def generate_signed_url(self, key, expiration=3600, use_external_host=True):
        rps = self._client.createSignedUrl("GET", self._bucket, key, expires=expiration)
        return rps.signedUrl

    def generate_upload_credentials(self, key, expiration=3600, use_external_host=True):
        content_type = 'text/plain'
        mime_type = mimetypes.guess_type(key)[0]
        if mime_type == 'text/plain':
            mime_type = 'text/plain;charset=utf-8'
        if mime_type:
            content_type = mime_type
        response = self._client.createSignedUrl("PUT", self._bucket, key,
                                                expires=expiration, headers={'Content-Type': content_type})
        return response



from abc import abstractmethod, ABC
from enum import Enum


class Protocol(Enum):
    OBS = 1
    BOTO3 = 2


class StorageInterface(ABC):

    @abstractmethod
    def list_objects(self, key, delimiter=''):
        pass

    @abstractmethod
    def get_object(self, key):
        pass

    @abstractmethod
    def put_object(self, key, content):
        pass

    @abstractmethod
    def upload_file(self, local_filename, key):
        pass

    @abstractmethod
    def upload_dir(self, local_dir, key):
        pass

    @abstractmethod
    def download_file(self, key, local_filename):
        pass

    @abstractmethod
    def download_dir(self, key, local_dir):
        pass

    @abstractmethod
    def copy_object(self, source_key, target_key):
        pass

    @abstractmethod
    def copy_dir(self, source_dir, dest_dir):
        pass

    @abstractmethod
    def create_dir(self, dirname):
        pass

    @abstractmethod
    def generate_signed_url(self, key, expiration=3600, use_external_host=True):
        pass

    @abstractmethod
    def generate_upload_credentials(self, key, expiration=3600, use_external_host=True):
        pass


def create_storage(protocol: Protocol, cfgs=None):
    if cfgs is None:
        cfgs = {}
    if protocol == Protocol.OBS:
        from taichu_storage.obs_client import StorageObs
        return StorageObs(cfgs)
    if protocol == Protocol.BOTO3:
        from taichu_storage.boto3_client import StorageBoto3
        return StorageBoto3(cfgs)
    return None




import logging
from io import BytesIO

import boto3
from botocore.config import Config
from taichu_storage import StorageInterface
import os


class StorageBoto3(StorageInterface):
    def __init__(self, cfgs=None):
        if cfgs is None:
            cfgs = {}
        ak = cfgs.get('ak')
        sk = cfgs.get('sk')
        endpoint_url = cfgs.get('endpoint_url')
        prefix = cfgs.get('bucket')
        endpoint_url_external = cfgs.get('endpoint_url_external')
        if prefix and not prefix.endswith('/'):
            prefix = prefix + '/'
        self._prefix = prefix
        self._bucket = 'alluxio'
        self._client = self.init_client(ak, sk, endpoint_url)
        if endpoint_url_external:
            self._client_external = self.init_client(ak, sk, endpoint_url_external)
        else:
            self._client_external = self._client

    def init_client(self, ak, sk, endpoint_url):
        client = boto3.client(
            's3',
            aws_access_key_id=ak,
            aws_secret_access_key=sk,
            use_ssl=True,
            region_name='default',
            endpoint_url=endpoint_url,
            config=Config(s3={"addressing_style": "path", "signature_version": 's3v4'},
                          connect_timeout=1, retries={'max_attempts': 1}))
        client.list_buckets()
        print(f"Successfully connected to S3, endpoint_url: {endpoint_url}")
        return client

    def list_objects(self, key, delimiter=''):
        key = f'{self._prefix}{key}'
        result = []
        response = self._client.list_objects_v2(Bucket=self._bucket, Prefix=key, Delimiter=delimiter)
        if 'CommonPrefixes' in response:
            folders = []
            for item in response['CommonPrefixes']:
                folders.append({
                    'name': item['Prefix'].replace(self._prefix, '', 1),
                    'is_dir': True,
                    'size': 0,
                    'last_modified': None,
                })
            result.extend(folders)
        if 'Contents' in response:
            objects = []
            for item in response['Contents']:
                if item['Key'].endswith('/'):
                    continue
                objects.append({
                    'name': item['Key'].replace(self._prefix, '', 1),
                    'is_dir': False,
                    'size': item['Size'],
                    'last_modified': item['LastModified'].strftime("%Y-%m-%d %H:%M:%S"),
                })
            result.extend(objects)
        return result

    def get_object(self, key):
        key = f'{self._prefix}{key}'
        response = self._client.get_object(Bucket=self._bucket, Key=key)
        return BytesIO(response.get('Body').read())

    def put_object(self, key, content):
        key = f'{self._prefix}{key}'
        self._client.put_object(Body=content, Bucket=self._bucket, Key=key)

    def upload_file(self, local_filename, key):
        key = f'{self._prefix}{key}'
        self._client.upload_file(local_filename, self._bucket, key)

    def upload_dir(self, local_dir, key):
        local_dir = local_dir.rstrip('/')
        key = key.rstrip('/')
        for item in os.scandir(local_dir):
            remote_key = item.path.replace(local_dir, key)
            if item.is_file():
                self.upload_file(item.path, remote_key)
            elif item.is_dir():
                self.upload_dir(item.path, remote_key)

    def download_file(self, key, local_filename):
        key = f'{self._prefix}{key}'
        self._client.download_file(self._bucket, key, local_filename)

    def download_dir(self, key, local_dir):
        key = f'{self._prefix}{key}'
        response = self._client.list_objects_v2(Bucket=self._bucket, Prefix=key)
        m = {}
        for obj in response.get('Contents', []):
            s3_key = obj['Key']
            local_file_path = f'{local_dir}{s3_key.replace(key, "")}'
            dirname = os.path.dirname(local_file_path)
            if not m.get(dirname, False):
                os.makedirs(dirname, exist_ok=True)
                m[dirname] = True
            if local_file_path.endswith('/'):
                continue
            self._client.download_file(self._bucket, s3_key, local_file_path)

    def copy_object(self, source_key, target_key):
        target_dir = os.path.dirname(target_key)
        self.create_dir(target_dir)
        source_key = f'{self._prefix}{source_key}'
        target_key = f'{self._prefix}{target_key}'
        self._client.copy_object(Bucket=self._bucket, Key=target_key,
                                 CopySource={'Bucket': self._bucket, 'Key': source_key})

    def copy_dir(self, source_dir, dest_dir):
        source_dir = f'{self._prefix}{source_dir}'
        dest_dir = f'{self._prefix}{dest_dir}'
        objects = self._client.list_objects_v2(Bucket=self._bucket, Prefix=source_dir)
        if 'Contents' not in objects:
            logging.info("No files found in the source folder.")
            return
        m = {}
        for obj in objects['Contents']:
            if obj['Key'].endswith('/'):
                continue
            copy_source = {'Bucket': self._bucket, 'Key': obj['Key']}
            dest_key = obj['Key'].replace(source_dir, dest_dir, 1)
            dirname = os.path.dirname(dest_key)
            if not m.get(dirname):
                self._client.put_object(Bucket=self._bucket, Key=dirname + '/')
                m[dirname] = True
            self._client.copy_object(CopySource=copy_source, Bucket=self._bucket, Key=dest_key)

    def create_dir(self, dirname):
        key = f'{self._prefix}{dirname}'
        if not key.endswith('/'):
            key = key + '/'
        self._client.put_object(Bucket=self._bucket, Key=key)

    def generate_signed_url(self, key, expiration=3600, use_external_host=True):
        key = f'{self._prefix}{key}'
        client = self._client
        if use_external_host:
            client = self._client_external
        url = client.generate_presigned_url(
            'get_object',
            Params={'Bucket': self._bucket, 'Key': key},
            ExpiresIn=expiration,
        )
        return url

    def generate_upload_credentials(self, key, expiration=3600, use_external_host=True):
        key = f'{self._prefix}{key}'
        client = self._client
        if use_external_host:
            client = self._client_external
        response = client.generate_presigned_post(self._bucket, key, ExpiresIn=expiration)

        return response



import argparse
import sys
from taichu_storage.obs_client import StorageObs


def cli():
    parser = argparse.ArgumentParser(description="""Taichu Storage Tools""")
    parser.add_argument('action', action="store", choices=['cp'])
    parser.add_argument('source', action="store",  type=str)
    parser.add_argument('destination', action="store",type=str)
    parser.add_argument('--ak', action="store",
                        help="""Specify the access key""")
    parser.add_argument('--sk', action="store", type=str,
                        help="""Specify the secret key""")
    parser.add_argument('--endpoint', action="store", type=str,
                        help="""Specify the endpoint""")

    args = parser.parse_args()

    if args.source is None or args.destination is None:
        print("Please specify both source and destination")
        sys.exit(1)

    print("Copying {} to {}".format(args.source, args.destination))
    if args.source.startswith("obs://") and not args.destination.startswith("obs://"):
        download_from_obs(args, args.source, args.destination)
    else:
        print("Not supported yet")


def download_from_obs(args, src, dest):
    src_bucket, src_key = parse_obs_path(src)
    client = StorageObs(cfgs={
        'bucket': src_bucket,
        'ak': args.ak,
        'sk': args.sk,
        'endpoint_url': args.endpoint})
    if src_key.endswith("/"):
        client.download_dir(src_key, dest)
    else:
        client.download_file(src_key, dest)


def parse_obs_path(obs_path):
    if obs_path.startswith("obs://"):
        obs_path = obs_path[6:]

    bucket = obs_path.split("/")[0]
    key = obs_path[len(bucket) + 1:]

    return bucket, key
